#ifndef TEST_BUILD

#define COMMAND_STRUCTURE_ERROR 1
#define INPUTFILE_ERROR 2
#define PATTERN_ERROR 3
#define NEGATIVE_PARMETER 4
#define EMPTY_NOTES 5

#include <imaging/bitmap.h>
#include <imaging/bmp-format.h>
#include <imaging/color.h>
#include <midi/primitives.h>
#include <midi/midi.h>
#include <shell/command-line-parser.h>
#include <Catch.h>
#include <fstream>
#include <iomanip>
#include <ios>
#include <iostream>
#include <sstream>
#include <string>

using namespace imaging;

Color findColor(const midi::Instrument& instrument)
{
	double x = static_cast<double>(value(instrument)) * 2 + 1;
	switch (value(instrument) % 6) {
		case 0:
			return Color{ 0, 0, x };
			break;
		case 1:
			return Color{ 0, x, 0 };
			break;
		case 2:
			return Color{ x, 0, 0 };
			break;
		case 3:
			return Color{ 0, x, x };
			break;
		case 4:
			return Color{ x, 0, x };
			break;
		default:
			return Color{ x, x, 0 };
	}
}

void drawRectangle(Bitmap& bitmap, const unsigned& width, const unsigned& height, const Position& position, const Color& color)
{
	for (int x = 0; x < width; x++)
	{
		for (int y = 0; y < height; y++)
		{
			bitmap[Position(position.x + x, position.y + y)] = color;
		}
	}
}

void drawBitmap(Bitmap& bitmap, const std::vector<midi::NOTE>& notes, const uint8_t& highest, const unsigned& pixelHeight, const unsigned& horizontalScale)
{
	for (midi::NOTE note : notes)
	{
		drawRectangle(bitmap, floor(value(note.duration) * static_cast<double>(horizontalScale) / 100), pixelHeight, Position(floor(value(note.start) * static_cast<double>(horizontalScale) / 100), (highest - value(note.note_number)) * pixelHeight), findColor(note.instrument));//kleuren[value(note.instrument) % kleuren.size()]);
	}
}

void makeBmpFiles(Bitmap& bitmap, const unsigned& frameWidth, const unsigned& totalWidth, const unsigned& totalHeight, const unsigned& horizontalStep, const unsigned& pixelHeight, const std::string& pattern)
{
	std::string prefix = pattern.substr(0, pattern.find("%d"));
	std::string suffix = pattern.substr(pattern.find("%d")+2, pattern.length()-1);
	if (totalWidth != frameWidth)
	{
		for (int i = 0; i < (totalWidth - frameWidth) / horizontalStep; i++)
		{
			std::stringstream ss;
			ss << std::setfill('0') << std::setw(5) << i;
			imaging::save_as_bmp(prefix + ss.str() + suffix, *bitmap.slice(i * horizontalStep, 0, frameWidth, totalHeight*pixelHeight));
		}
	}
	else
	{
		imaging::save_as_bmp(prefix + "00000" + suffix, bitmap);
	}

}

int analyzeNotes(const std::vector<midi::NOTE>& notes, uint8_t& highest, uint8_t& lowest, unsigned& width)
{
	if (notes.size() != 0)
	{
		highest = value(notes[0].note_number);
		lowest = value(notes[0].note_number);
		width = value(notes[0].start + notes[0].duration);
		for (int i = 1; i < notes.size(); i++)
		{
			uint8_t note = value(notes[i].note_number);
			if (note > highest)
			{
				highest = note;
			}
			if (note < lowest)
			{
				lowest = note;
			}
			if (value(notes[i].start + notes[i].duration) > width)
			{
				width = value(notes[i].start + notes[i].duration);
			}
		}
		return 0;
	}
	else
	{
		std::cout << "the given file doesn't have notes" << std::endl;
		return EMPTY_NOTES;
	}
}

int main(int argn, char** argv)
{
	unsigned totalWidth;
	unsigned frameWidth = 0;
	unsigned totalHeight;
	unsigned horizontalStep = 1;
	unsigned horizontalScale = 5;
	unsigned pixelHeight = 16;
	std::string filename;//must be specified
	std::string pattern;//must be specified

	shell::CommandLineParser parser;
	parser.add_argument("-w", &frameWidth);
	parser.add_argument("-d", &horizontalStep);
	parser.add_argument("-s", &horizontalScale);
	parser.add_argument("-h", &pixelHeight);
	parser.process(argn, argv);

	std::vector<std::string> positional_arguments = parser.positional_arguments();
	if (positional_arguments.size() == 2 && positional_arguments[1].find("%d"))
	{
		filename = positional_arguments[0];
		pattern = positional_arguments[1];
	}
	else
	{
		//input and output file check
		if (positional_arguments.size() == 0)
		{
			std::cout << "no filename or pattern given" << std::endl;
		}
		else if (positional_arguments.size() == 1)
		{
			if (positional_arguments[0].find(".mid"))
			{
				std::cout << "patern missing" << std::endl;
				return PATTERN_ERROR;
			}
			else
			{
				std::cout << "input file missing" << std::endl;
				return INPUTFILE_ERROR;
			}
		}
		else if (positional_arguments.size() == 2)
		{
			std::cout << "'%d' must be in pattern" << std::endl;
			return PATTERN_ERROR;
		}
		else
		{
			std::cout << "too many positional arguments given" << std::endl;
		}
		return COMMAND_STRUCTURE_ERROR;
	}

	std::ifstream inputfile (filename, std::ios::binary);
	std::vector<midi::NOTE> notes = midi::read_notes(inputfile);
	
	uint8_t highest, lowest;
	int exitCode = analyzeNotes(notes, highest, lowest, totalWidth);
	if (exitCode != 0)
	{
		return exitCode;
	}

	if (frameWidth == 0 || frameWidth > floor(totalWidth * static_cast<double>(horizontalScale) / 100))
	{
		frameWidth = floor(totalWidth * static_cast<double>(horizontalScale) / 100);
	}
	totalHeight = highest - lowest + 1;

	//parameter check
	if (frameWidth < 0)
	{
		std::cout << "The frame width must be a positive number" << std::endl;
		return NEGATIVE_PARMETER;
	}
	else if (horizontalStep < 0)
	{
		std::cout << "The horizontal step must be a positive number" << std::endl;
		return NEGATIVE_PARMETER;
	}
	else if (horizontalScale < 0)
	{
		std::cout << "The horizontal scale must be a positive number" << std::endl;
		return NEGATIVE_PARMETER;
	}
	else if (pixelHeight < 0)
	{
		std::cout << "The pixel height must be a positive number" << std::endl;
		return NEGATIVE_PARMETER;
	}

	Bitmap bitmap = Bitmap(floor(totalWidth * static_cast<double>(horizontalScale) / 100), totalHeight*pixelHeight);
	drawBitmap(bitmap, notes, highest, pixelHeight, horizontalScale);
	makeBmpFiles(bitmap, frameWidth, floor(totalWidth * static_cast<double>(horizontalScale) / 100), totalHeight, horizontalStep, pixelHeight, pattern);
}

#endif
