#include "midi.h"
#include "midi/primitives.h"
#include "io/read.h"
#include "io/endianness.h"
#include "io\vli.h"
#include <iostream>
#include <string>
#include <io.h>
using namespace std;

namespace midi 
{
    void read_chunk_header(istream& ss, CHUNK_HEADER* header)
    {
        io::read_to(ss, &header->id);

        io::read_to(ss, &header->size);
        io::switch_endianness(&header->size);
    }

    string header_id(CHUNK_HEADER header)
    {
        return string(header.id, 4);
    }

    void read_mthd(std::istream& ss, MTHD* mthd)
    {
        read_chunk_header(ss, &mthd->header);

        io::read_to(ss, &mthd->type);
        io::switch_endianness(&mthd->type);

        io::read_to(ss, &mthd->ntracks);
        io::switch_endianness(&mthd->ntracks);

        io::read_to(ss, &mthd->division);
        io::switch_endianness(&mthd->division);
    }

    bool is_sysex_event(uint8_t byte)
    {
        return byte == 0xF0 || byte == 0xF7;
    }

    bool is_meta_event(uint8_t byte)
    {
        return byte == 0xFF;
    }

    bool is_midi_event(uint8_t byte)
    {
        return byte >= 0x80 && byte <= 0xEF;
    }

    bool is_running_status(uint8_t byte)
    {
        return byte <= 0x7F;
    }

    uint8_t extract_midi_event_type(uint8_t status)
    {
        return status >> 4;
    }

    Channel extract_midi_event_channel(uint8_t status)
    {
        status <<= 4;
        return Channel(status >> 4);
    }

    bool is_note_off(uint8_t status)
    {
        return status == 0x08;
    }

    bool is_note_on(uint8_t status)
    {
        return status == 0x09;
    }

    bool is_polyphonic_key_pressure(uint8_t status)
    {
        return status == 0x0A;
    }

    bool is_control_change(uint8_t status)
    {
        return status == 0x0B;
    }

    bool is_program_change(uint8_t status)
    {
        return status == 0x0C;
    }

    bool is_channel_pressure(uint8_t status)
    {
        return status == 0x0D;
    }

    bool is_pitch_wheel_change(uint8_t status)
    {
        return status == 0x0E;
    }

    void read_mtrk(std::istream& in, EventReceiver& receiver)
    {
        CHUNK_HEADER header;
        read_chunk_header(in, &header);

        if (header_id(header) == "MTrk") {

            bool end_of_track_reached = false;
            Duration dt;
            uint8_t identifier{};
            uint8_t running_status = false;

            while (!end_of_track_reached)
            {
                dt = Duration(io::read_variable_length_integer(in));
                if (!is_running_status(in.peek())) {
                    identifier = io::read<uint8_t>(in);
                }

                if (is_meta_event(identifier))
                {
                    uint8_t type = io::read<uint8_t>(in);
                    uint64_t length = io::read_variable_length_integer(in);

                    if (type == 0x2F && length == 0)
                    {
                        receiver.meta(dt, type, 0x00, 0);
                        end_of_track_reached = true;
                    }
                    else
                    {
                        auto data = io::read_array<uint8_t>(in, length);
                        receiver.meta(dt, type, std::move(data), length);

                    }
                }
                else if (is_sysex_event(identifier))
                {
                    uint64_t length = io::read_variable_length_integer(in);
                    auto data = io::read_array<uint8_t>(in, length);

                    receiver.sysex(dt, std::move(data), length);
                }
                else {
                    uint8_t type = extract_midi_event_type(identifier);
                    Channel channel = extract_midi_event_channel(identifier);

                    if (is_note_off(type))
                    {
                        NoteNumber note = NoteNumber(io::read<uint8_t>(in));
                        uint8_t velocity = io::read<uint8_t>(in);
                        receiver.note_off(dt, channel, note, velocity);
                    }
                    else if (is_note_on(type))
                    {
                        NoteNumber note = NoteNumber(io::read<uint8_t>(in));
                        uint8_t velocity = io::read<uint8_t>(in);
                        receiver.note_on(dt, channel, note, velocity);
                    }
                    else if (is_polyphonic_key_pressure(type))
                    {
                        NoteNumber note = NoteNumber(io::read<uint8_t>(in));
                        uint8_t pressure = io::read<uint8_t>(in);
                        receiver.polyphonic_key_pressure(dt, channel, note, pressure);
                    }
                    else if (is_control_change(type))
                    {
                        uint8_t controller = io::read<uint8_t>(in);
                        uint8_t value = io::read<uint8_t>(in);
                        receiver.control_change(dt, channel, controller, value);
                    }
                    else if (is_program_change(type))
                    {
                        Instrument program = Instrument(io::read<uint8_t>(in));
                        receiver.program_change(dt, channel, program);
                    }
                    else if (is_channel_pressure(type))
                    {
                        uint8_t pressure = io::read<uint8_t>(in);
                        receiver.channel_pressure(dt, channel, pressure);
                    }
                    else if (is_pitch_wheel_change(type))
                    {
                        uint16_t value = io::read<uint8_t>(in);
                        value |= (io::read<uint8_t>(in) << 7);
                        receiver.pitch_wheel_change(dt, channel, value);
                    }
                }
            }
        }
    }

    bool operator ==(const NOTE& note1, const NOTE& note2)
    {
        return (note1.note_number == note2.note_number) && (note1.start == note2.start) && (note1.duration == note2.duration) && (note1.velocity == note2.velocity) && (note1.instrument == note2.instrument);
    }

    bool operator !=(const NOTE& note1, const NOTE& note2)
    {
        return !(note1 == note2);
    }

    std::ostream& operator<<(std::ostream& out, const NOTE& note)
    {
        return out << "Note(number=" << note.note_number << ",start=" << note.start << ",duration=" << note.duration << ",instrument=" << note.instrument << ")";
    }

// ChannelNoteCollector

    void ChannelNoteCollector::note_off(Duration dt, Channel channel, NoteNumber note, uint8_t velocity)
    {
        start += dt;
        if (this->channel == channel && active_notes.count(note))
        {
            NOTE active_note = active_notes[note];
            active_note.duration = start - active_note.start;
            note_receiver(active_note);
            active_notes.erase(note);
        }
    }
    void ChannelNoteCollector::note_on(Duration dt, Channel channel, NoteNumber note, uint8_t velocity)
    {
        if (this->channel == channel)
        {
            if (velocity != 0) {
                if (active_notes.count(note)) {
                    note_off(dt, channel, note, velocity);
                }
                else
                {
                    start += dt;
                }
                active_notes.insert(std::pair<NoteNumber, NOTE>(note, NOTE(note, start, Duration(), velocity, this->program)));
            }
            else
            {
                note_off(dt, channel, note, velocity);
            }
        }
        else
        {
            start += dt;
        }
    }
    void ChannelNoteCollector::polyphonic_key_pressure(Duration dt, Channel channel, NoteNumber note, uint8_t pressure)
    {
        start += dt;
    }
    void ChannelNoteCollector::control_change(Duration dt, Channel channel, uint8_t controller, uint8_t value)
    {
        start += dt;
    }
    void ChannelNoteCollector::program_change(Duration dt, Channel channel, Instrument program)
    {
        start += dt;
        if (this->channel == channel)
        {
            this->program = program;
        }
    }
    void ChannelNoteCollector::channel_pressure(Duration dt, Channel channel, uint8_t pressure)
    {
        start += dt;
    }
    void ChannelNoteCollector::pitch_wheel_change(Duration dt, Channel channel, uint16_t value)
    {
        start += dt;
    }

    void ChannelNoteCollector::meta(Duration dt, uint8_t type, std::unique_ptr<uint8_t[]> data, uint64_t data_size)
    {
        start += dt;
    }
    void ChannelNoteCollector::sysex(Duration dt, std::unique_ptr<uint8_t[]> data, uint64_t data_size)
    {
        start += dt;
    }

// EventMulticaster
    void EventMulticaster::note_off(Duration dt, Channel channel, NoteNumber note, uint8_t velocity)
    {
        for (int i = 0; i < eventReceivers.size(); i++)
        {
            eventReceivers[i]->note_off(dt, channel, note, velocity);
        }
    }
    void EventMulticaster::note_on(Duration dt, Channel channel, NoteNumber note, uint8_t velocity)
    {
        for (int i = 0; i < eventReceivers.size(); i++)
        {
            eventReceivers[i]->note_on(dt, channel, note, velocity);
        }
    }
    void EventMulticaster::polyphonic_key_pressure(Duration dt, Channel channel, NoteNumber note, uint8_t pressure)
    {
        for (int i = 0; i < eventReceivers.size(); i++)
        {
            eventReceivers[i]->polyphonic_key_pressure(dt, channel, note, pressure);
        }
    }
    void EventMulticaster::control_change(Duration dt, Channel channel, uint8_t controller, uint8_t value)
    {
        for (int i = 0; i < eventReceivers.size(); i++)
        {
            eventReceivers[i]->control_change(dt, channel, controller, value);
        }
    }
    void EventMulticaster::program_change(Duration dt, Channel channel, Instrument program)
    {
        for (int i = 0; i < eventReceivers.size(); i++)
        {
            eventReceivers[i]->program_change(dt, channel, program);
        }
    }
    void EventMulticaster::channel_pressure(Duration dt, Channel channel, uint8_t pressure)
    {
        for (int i = 0; i < eventReceivers.size(); i++)
        {
            eventReceivers[i]->channel_pressure(dt, channel, pressure);
        }
    }
    void EventMulticaster::pitch_wheel_change(Duration dt, Channel channel, uint16_t value)
    {
        for (int i = 0; i < eventReceivers.size(); i++)
        {
            eventReceivers[i]->pitch_wheel_change(dt, channel, value);
        }
    }
    void EventMulticaster::meta(Duration dt, uint8_t type, std::unique_ptr<uint8_t[]> data, uint64_t data_size)
    {
        for (int i = 0; i < eventReceivers.size(); i++)
        {
            eventReceivers[i]->meta(dt, type, std::move(data), data_size);
        }
    }
    void EventMulticaster::sysex(Duration dt, std::unique_ptr<uint8_t[]> data, uint64_t data_size)
    {
        for (int i = 0; i < eventReceivers.size(); i++)
        {
            eventReceivers[i]->sysex(dt, std::move(data), data_size);
        }
    }

// NoteCollector

    void NoteCollector::note_off(Duration dt, Channel channel, NoteNumber note, uint8_t velocity)
    {
        eventMulticaster.note_off(dt, channel, note, velocity);
    }
    void NoteCollector::note_on(Duration dt, Channel channel, NoteNumber note, uint8_t velocity)
    {
        eventMulticaster.note_on(dt, channel, note, velocity);
    }
    void NoteCollector::polyphonic_key_pressure(Duration dt, Channel channel, NoteNumber note, uint8_t pressure)
    {
        eventMulticaster.polyphonic_key_pressure(dt, channel, note, pressure);
    }
    void NoteCollector::control_change(Duration dt, Channel channel, uint8_t controller, uint8_t value)
    {
        eventMulticaster.control_change(dt, channel, controller, value);
    }
    void NoteCollector::program_change(Duration dt, Channel channel, Instrument program)
    {
        eventMulticaster.program_change(dt, channel, program);
    }
    void NoteCollector::channel_pressure(Duration dt, Channel channel, uint8_t pressure)
    {
        eventMulticaster.channel_pressure(dt, channel, pressure);
    }
    void NoteCollector::pitch_wheel_change(Duration dt, Channel channel, uint16_t value)
    {
        eventMulticaster.channel_pressure(dt, channel, value);
    }
    void NoteCollector::meta(Duration dt, uint8_t type, std::unique_ptr<uint8_t[]> data, uint64_t data_size)
    {
        eventMulticaster.meta(dt, type, std::move(data), data_size);
    }
    void NoteCollector::sysex(Duration dt, std::unique_ptr<uint8_t[]> data, uint64_t data_size)
    {
        eventMulticaster.sysex(dt, std::move(data), data_size);
    }

    std::vector<NOTE> read_notes(std::istream& in)
    {
        MTHD mthd_chunk;
        read_mthd(in, &mthd_chunk);
        std::vector<NOTE> notes {};
        if (mthd_chunk.ntracks != 0) {
            for (int i = 0; i < mthd_chunk.ntracks; i++)
            {
                NoteCollector collector([&notes](const midi::NOTE& note) { notes.push_back(note); });
                read_mtrk(in, collector);
            }
        }
        return notes;
    }
}